import os
import csv
import torch
import cv2 as cv
import numpy as np
import pandas as pd
from tqdm.notebook import tqdm
from torchvision import utils
from matplotlib import pyplot as plt
from jai.dataset import JaiDataset, jai_split


def extract_img_from_parquet(src_dir, dst_dir):
    """
    Extract the images from the parquet files.
    Apply to train/test data separately.
    :param src_dir: src folder contains parquet images.
    :param dst_dir: dst folder for placing the extracted images. Make sure it is clean.
    :return: void
    """

    curr_folder = os.path.join(dst_dir, "0")
    os.mkdir(curr_folder)
    for file in sorted(os.listdir(src_dir)):
        if file.endswith('.parquet'):
            df = pd.read_parquet(os.path.join(src_dir, file))
            print("Total of {} images in {}.".format(len(df), os.path.basename(file)))
            for i in tqdm(range(len(df))):
                img_id = df.iat[i, 0]
                img_idx = int(img_id[img_id.find('_') + 1:])
                img = df.iloc[i, 1:].to_numpy().reshape((137, 236)).astype(np.uint8)

                if img_idx % 10000 == 0 and img_idx > 0:
                    print("Processed {} images".format(i))
                    next_folder = os.path.join(dst_dir, "{}").format(img_idx // 10000)
                    os.mkdir(next_folder)
                    curr_folder = next_folder

                file_name = os.path.join(curr_folder, "{}.png".format(img_id))
                cv.imwrite(file_name, img, [cv.IMWRITE_PNG_COMPRESSION, 0])


class BengaliParquetData(JaiDataset):
    """
    Dataset Class (Test Time) for Bengali Parquet Data.
    TODO: Need to test this on server
    """

    HEIGHT = 137
    WIDTH = 236

    def __init__(self, fp, tsfms, usage, ann_src=None, augmentators=None):
        """
        Constructor
        :param fp: file path(s), if a directory was provided, will read all parquet files from the directory
        :param tsfms: image preprocessing transformers
        """
        super().__init__(augmentators)
        self.ann = []
        if usage == 'test':
            pass
        elif usage == 'train' and ann_src is not None:
            # parse annotation
            with open(file=ann_src) as csv_file:
                reader = list(csv.DictReader(f=csv_file, fieldnames=['img_id', 'root', 'vowel', 'consonant',
                                                                     'grapheme']))
                row = 0
                for entry in reader:
                    if row > 0:
                        self.ann.append(entry)
                    row += 1
        else:
            raise Exception("Usage must be either train or eval. Annotation must be provided if the usage is train.")

        df = None
        for file in sorted(os.listdir(fp)):
            if file.startswith(usage) and file.endswith('parquet'):
                df_p = os.path.join(fp, file)
                df = pd.read_parquet(df_p) if df is None else df.append(pd.read_parquet(df_p), ignore_index=True)
                print("File {} loaded.".format(file))

        self.tsfms = tsfms
        self.usage = usage
        self.data_id = df.iloc[:, 0].values
        self.img_data = df.iloc[:, 1:].values.reshape(-1, self.HEIGHT, self.WIDTH).astype(np.uint8)

    def __len__(self):
        return len(self.img_data)

    def __getitem__(self, idx):

        img_id = self.data_id.iloc[idx, 0]
        img = self.img_data[idx]
        y = None

        if self.usage == 'train':
            annotations = self.ann[idx]
            if img_id != annotations['img_id']:
                raise Exception("Image ID and idx did not match!")

            root = int(annotations['root'])
            vowel = int(annotations['vowel'])
            consonant = int(annotations['consonant'])
            y = torch.Tensor([root, vowel, consonant]).int()

        for tsf in self.tsfms:
            img = tsf(img)

        img = self.augment(img)

        return {"id": img_id, "x": img, "y": y}


class BengaliLocalDataset(JaiDataset):
    """
    Dataset Class for Bengali Local Data
    """

    def __init__(self, img_src, ann_src, tsfms, augmentators=None):
        """
        Constructor
        :param img_src: img src folder
        :param ann_src: annotation src file csv
        :param tsfms: transformation
        :param augmentators: augmentations
        """

        super().__init__(augmentators)

        # input image size
        self.tsfms = tsfms

        # field
        self.img_src = img_src
        self.ann = []

        # parse annotation
        with open(file=ann_src) as csv_file:
            reader = list(csv.DictReader(f=csv_file, fieldnames=['img_id', 'root', 'vowel', 'consonant', 'grapheme']))
            row = 0
            for entry in reader:
                if row > 0:
                    self.ann.append(entry)
                row += 1

    def __getitem__(self, idx):
        """
        Get an image with annotations
        :param idx: idx of the item
        :return: data entry
        """

        if isinstance(idx, str):
            idx = int(idx[idx.find('_')+1:])

        # get annotation
        annotations = self.ann[idx]
        img_id = annotations['img_id']
        root = int(annotations['root'])
        vowel = int(annotations['vowel'])
        consonant = int(annotations['consonant'])
        y = torch.Tensor([root, vowel, consonant]).int()

        # gray scale
        img_idx = int(img_id[img_id.find('_')+1:])
        f_idx = int(img_idx // 1e4)
        img_p = os.path.join(self.img_src, "{}".format(f_idx), "{}.png".format(img_id))
        img = cv.imread(img_p, cv.IMREAD_GRAYSCALE)

        for tsf in self.tsfms:
            img = tsf(img)

        img = self.augment(img)

        return {"id": img_id, "x": img, "y": y}

    def __len__(self):
        return len(self.ann)

    def view_batch(self, n_row, img_ids=None):
        """
        View a random batch (12) of images
        :param n_row: number of samples = n_row * n_row
        :param img_ids: if provide a list of image id (len = n_row * n_row), show those
        :return: void
        """

        if img_ids is None:
            batch_idxes = np.random.randint(0, self.__len__(), n_row*n_row, dtype=np.int)
        elif len(img_ids) == n_row*n_row:
            batch_idxes = img_ids
        else:
            raise Exception("Length of img_ids need to be same as n_row*n_row")

        batch_processed = []
        batch_org = []

        for idx in batch_idxes:
            # get processed image
            batch_processed.append(self[idx]['x'])

            # get original image (in gray scale)
            img_id = self[idx]['id']
            img_idx = int(img_id[img_id.find('_') + 1:])
            f_idx = int(img_idx // 1e4)
            img_p = os.path.join(self.img_src, "{}".format(f_idx), "{}.png".format(img_id))
            org_img = (255 - cv.imread(img_p, cv.IMREAD_GRAYSCALE)).astype(np.float32) / 255
            batch_org.append(torch.from_numpy(org_img)[None, :, :])

        grid_pro = utils.make_grid(batch_processed, nrow=n_row)
        grid_org = utils.make_grid(batch_org, nrow=n_row)
        plt.title("Processed Image")
        plt.axis('off')
        plt.ioff()
        plt.imshow(grid_pro.numpy().transpose(1, 2, 0))
        plt.show()
        plt.title("Original Image (Reversed)")
        plt.axis('off')
        plt.ioff()
        plt.imshow(grid_org.numpy().transpose(1, 2, 0))
        plt.show()



from torch import randperm
from torch.utils.data import Dataset


def _accumulate(iterable, fn=lambda x, y: x + y):
    """Taken from PyTorch"""

    # _accumulate([1,2,3,4,5]) --> 1 3 6 10 15
    # _accumulate([1,2,3,4,5], operator.mul) --> 1 2 6 24 120
    it = iter(iterable)
    try:
        total = next(it)
    except StopIteration:
        return
    yield total
    for element in it:
        total = fn(total, element)
        yield total


def jai_split(dataset, lengths):
    """
    Random split for JaiDataset
    :param dataset: input dataset
    :param lengths: lengths
    :return: splitted dataset
    """

    if sum(lengths) != len(dataset):
        raise ValueError("Sum of input lengths does not equal the length of the input dataset!")

    indices = randperm(sum(lengths)).tolist()
    return [JaiSubset(dataset, indices[offset - length:offset], dataset.augmentators) for offset, length in zip(
        _accumulate(lengths), lengths)]


class JaiDataset(Dataset):
    """
    TODO: refactor the id access and string ID access to this parent class
    Abstract class for Jai Dataset that supports augmentation
    """

    def __init__(self, augments=None, *args):
        self.augmentators = augments

    def __getitem__(self, item):
        raise NotImplemented("")

    def __len__(self):
        raise NotImplemented("")

    def augment(self, img):
        """
        Don't forget to call this before getitem.
        :param img:
        :return:
        """

        if self.augmentators is not None:
            # For augment in self.augments:

            if isinstance(self.augmentators, list):
                for aug in self.augmentators:
                    img = aug.proc(img)
                return img
            else:
                return self.augmentators.proc(img)

    def split(self, train_ratio=0.8):
        """
        Split the dataset into two subset: train and eval
        :param train_ratio: train ratio
        :return: train eval sub set
        """

        if train_ratio <= 0 or train_ratio >= 1:
            raise Exception("Train ratio should be between 0 ~ 1!")

        train_len = int(self.__len__()*train_ratio)
        eval_len = self.__len__() - train_len
        lengths = [train_len, eval_len]
        train_set, eval_set = jai_split(self, lengths)
        return {'train': train_set, 'eval': eval_set}


class JaiSubset(JaiDataset):
    """
    Jai-style Subset
    TODO: refactor the string ID access to make it compatible with parent class
    """

    def __init__(self, dataset, indices, augments=None):
        super().__init__(augments)
        self.dataset = dataset
        self.indices = indices

    def __getitem__(self, idx):
        return self.dataset[self.indices[idx]]

    def __len__(self):
        return len(self.indices)


class DataClassDict:
    """
    Dummy class to cache the encoding
    """

    def __init__(self, names: list or str, n_classes: list or int):
        """
        Constructor. Input category order should match with the model output.
        :param names: prediction names
        :param n_classes: number of classes for the prediction
        :return: void
        """

        assert0 = isinstance(names, str) and isinstance(n_classes, int)
        if assert0:
            names = [names]
            n_classes = [n_classes]
        assert1 = isinstance(names, list) and isinstance(n_classes, list) and len(names) == len(n_classes)

        assert assert1, "names and n_classes do not match!"

        self.names = names
        self.n_classes = n_classes

    def items(self):
        return zip(self.names, self.n_classes)




"""
Image processing methods.
Mainly for preprocessing.
"""

import cv2 as cv
import numpy as np
import torch


def reverse_color(img):
    """
    Inverse the color
    :param img: img
    :return: color reversed img
    """

    if img.dtype != np.uint8:
        raise Exception("Input image should be in np.uint8 but was {}".format(img.dtype))

    return 255 - img


def grayscale(img, order='BGR'):
    """
    Gray scale the image to single channel
    :param img: image
    :param order: input image order: BGR for cv.imread amd RBG for PIL
    :return: gray-scaled image
    """

    if order == 'BGR':
        return cv.cvtColor(img, cv.COLOR_BGR2GRAY)
    elif order == 'RBG':
        return cv.cvtColor(img, cv.COLOR_RGB2GRAY)
    else:
        raise Exception("Channel order must be either BGR (opencv) or RBG (PIL) but was {}.".format(order))


def denoise(img, ksize=(3, 3)):
    """
    Denoise the image
    :param img: image
    :param ksize: kernel size for smoothing
    :return: denoised image
    """

    img = cv.blur(img, ksize=ksize)
    return img


def threshold(img, low=0, adaptive_ksize=None, C=None, binary=False):
    """
    Adaptive threshold only work for single channel image
    Threshold the image. Hard threshold or/and adaptive threshold.
    Adaptive threshold will use the mean of the kernal - C
    :param img: image
    :param low: if not None pixel below the lower bound will be set to 0
    :param adaptive_ksize: if not none, apply adaptive threshold
    :param C: value that will be subtracted from the adaptive kernel mean
    :param binary: whether return binary output
    :return: thresholded image
    """

    if low > 0:
        img[img < low] = 0
    mask = None
    if adaptive_ksize:
        mask = cv.adaptiveThreshold(img, maxValue=255, adaptiveMethod=cv.ADAPTIVE_THRESH_MEAN_C,
                                    thresholdType=cv.THRESH_BINARY, blockSize=13, C=C)
    if mask is not None:
        if binary:
            return mask
        else:
            return img * (mask == 255)
    return img


def centralize_object(img):
    """
    Translate the image to the pixel center
    Works for 0 background images after the threshold. If not thresholded, this method does not have effect.
    :param img: image
    :return: centralized image
    """

    height, width = img.shape[0], img.shape[1]
    col_map = np.repeat(np.arange(0, width)[np.newaxis, :], height, axis=0)
    row_map = np.repeat(np.arange(0, height)[:, np.newaxis], width, axis=1)
    row_centroid = row_map[img != 0].mean()  # get the nonzero pixels across all rows and find the mean
    col_centroid = col_map[img != 0].mean()  # get the nonzero pixels across all cols and find the mean
    row_t = height / 2 - row_centroid
    col_t = width / 2 - col_centroid
    T = np.array([[1, 0, col_t], [0, 1, row_t]])  # translate to the pixel center
    img = cv.warpAffine(img, T, (width, height))
    return img


def resize_object(img, size, pad, margin=5):
    """
    Crop the image using the the non-zero boundary (object), pad and resize
    Code borrowed from https://www.kaggle.com/iafoss/image-preprocessing-128x128
    :param img: image
    :param pad: padding around the bounding box
    :param size: new size
    :param margin: margin of the bounding box around the object
    :return:
    """

    # get the valid range (bbox for the writing) along two axes
    mask = img != 0
    rows = np.any(mask, axis=1)
    cols = np.any(mask, axis=0)
    rmin, rmax = np.where(rows)[0][[0, -1]]
    cmin, cmax = np.where(cols)[0][[0, -1]]

    # crop the bbox with some additional margins
    cmin = cmin - margin if (cmin > margin) else 0
    rmin = rmin - margin if (rmin > margin) else 0
    cmax = cmax + margin if (cmax < img.shape[1] - margin) else img.shape[1]
    rmax = rmax + margin if (rmax < img.shape[0] - margin) else img.shape[0]
    img = img[rmin:rmax, cmin:cmax]

    # cropped size
    lc, lr = cmax - cmin, rmax - rmin
    # find longer edge
    ledge = max(lc, lr) + pad
    # make sure that the aspect ratio is kept in rescaling
    img = np.pad(img, [((ledge - lr) // 2,), ((ledge - lc) // 2,)], mode='constant')
    return cv.resize(img, (size, size))


def rescale(img, dtype=np.float32):
    """
    rescale image to 0-1
    :param img: image
    :param dtype: data type for the ouput
    :return: normalized image
    """

    img = (img / (img.max() + 1e-8)).astype(dtype)
    return img


def standardize(img, dtype=np.float32):
    """
    standardize the data to 0 mean and unit var
    :param img: image
    :param dtype: output datatype
    :return: standardized data
    """

    img -= img.mean()
    img = img / (img.std() + 1e-8)
    img.astype(dtype)

    return img


def add_bgnoise(mu, sigma, img):
    """
    Add some background noise to the binary image
    This will convert image to float32
    :param mu: mean
    :param sigma: std
    :param img: image
    :return:
    """

    if img.dtype == np.uint8:
        img = img.astype(np.float32) / 255
    gauss_noise = np.random.normal(mu, sigma, img.shape[0]*img.shape[1]).reshape(img.shape)
    img += gauss_noise
    img[img > 1.0] = 1.0
    img[img < 0] = 0
    return img


def to_tensor(img, dtype=torch.float32):
    """
    To tensor in float
    :param img: image
    :param dtype: output data type
    :return: image in tensor and the
    """

    if img.dtype == np.uint8:
        img = img.astype(np.float32) / 255
    if img.ndim == 2:  # if the input image is grayscale, add the pseudo dimension
        img = torch.from_numpy(img)[None, :, :].to(dtype)  # HxW -> 1xHxW
    elif img.ndim == 3:
        img = torch.from_numpy(img).permute((2, 0, 1)).to(dtype)  # HxWx3 -> 3xHxW

    return img




"""Setup for the jai."""

import setuptools


with open('README.md', encoding='utf-8') as f:
    README = f.read()

requirements = []
desc = "jai - A LEGO-style & PyTorch-based Deep Learning Library"
with open('requirements.txt', mode='r') as req:
    reader = req.readlines()
    for pkg in reader:
        pkg = pkg.replace('=', '>=', 1)
        pkg = pkg[:pkg.rfind('=')]
        requirements.append(pkg)

setuptools.setup(
    author="Jia Geng",
    author_email="gengjia0214@hotmail.com",
    name='jai',
    license="BSD 3-Clause License",
    description=desc,
    version='v0.0.7',
    long_description=README,
    long_description_content_type='text/markdown',
    url='https://github.com/gengjia0214/jai.git',
    packages=setuptools.find_packages(),
    python_requires=">=3.6",
    install_requires=requirements,
    classifiers=[
        # Trove classifiers
        # (https://pypi.python.org/pypi?%3Aaction=list_classifiers)
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research'
    ],
)